package gui;

import java.io.File;
import java.util.ArrayList;

import org.opencv.core.Mat;
import dataManagement.DataManager;
import dataManagement.Page;
import geometry.PointList;
import importExport.PageXMLWriterOld;
import importExport.SegResultImport;
import regions.RegionManager;
import segmentation.ImageProcessor;
import segmentation.Segmenter;
import segmentation.parameters.Parameters;
import segmentation.result.SegmentationResult;

public class GuiManager implements Runnable {

	private boolean keepSegmenting;

	private String outputFolder;

	private Gui gui;
	private DataManager dataManager;
	private RegionManager regionManager;
	private Page currentPage;

	private Thread runner;

	public GuiManager(Gui gui) {
		setGui(gui);
		setDataManager(gui.getDataManager());
		setRegionManager(new RegionManager(gui));

		setCurrentPage(gui.getDataManager().getCurrentPage());
		gui.setGuiManager(this);
	}

	public void initFirstSegResult() {
		SegmentationResult segResult = loadSegResult(currentPage);

		if (segResult != null) {
			currentPage.setSegmentationResult(segResult);
			currentPage.visualizeSegmentation();
			gui.getRbPostProc().setEnabled(true);
			gui.getRbPostProc().doClick();

			return;
		}
	}
	
	public SegmentationResult loadSegResult(Page page) {
		String xmlPath = page.getImagePath().substring(0, page.getImagePath().lastIndexOf(".") + 1) + "xml";
		SegmentationResult segResult = null;
		
		if(new File(xmlPath).exists()) {
			segResult = SegResultImport.readPageXML(xmlPath, page);
		}

		return segResult;
	}

	public void saveCurrentPage(boolean tempResult) {
		if (currentPage.getSegmentationResult() != null) {
			PageXMLWriterOld.writePageXML(currentPage, outputFolder, tempResult);
		}
	}

	public void segmentAllPages() {
		setKeepSegmenting(true);
		gui.getPnlPages().getProcessedPages().removeAll(gui.getPnlPages().getProcessedPages());
		gui.getPnlPages().getListPages().repaint();

		segmentCurrentPage();
		startSegmentation();
	}

	public void segmentCurrentPage() {
		Mat original = currentPage.getOriginal();
		Parameters parameters = new Parameters(regionManager, gui);
		currentPage.setScaleFactor((double) currentPage.getOriginal().height() / parameters.getDesiredImageHeight());

		Segmenter segmenter = new Segmenter(parameters);
		SegmentationResult segmentationResult = segmenter.segment(original);
		currentPage.setSegmentationResult(segmentationResult);

		currentPage.visualizeSegmentation();
		gui.showImage(currentPage.getResult());
		// gui.getRbPostProc().setSelected(false);
		gui.getRbResult().setEnabled(true);
		gui.getRbPostProc().setEnabled(true);
		gui.getRbResult().doClick();
	}

	public void showNextPage() {
		int index = dataManager.getIndex();

		if (index < dataManager.getPages().size() - 1) {
			index++;
			gui.getPnlPages().getListPages().setSelectedIndex(index);
			// loadPage(dataManager.getPages().get(index).getFileName());
		}
	}

	public void showPreviousPage() {
		int index = dataManager.getIndex();

		if (index > 0) {
			index--;
			gui.getPnlPages().getListPages().setSelectedIndex(index);
			// loadPage(dataManager.getPages().get(index).getFileName());
		}
	}

	public Page identifyPage(String fileName) {
		Page page = null;

		for (int i = 0; i < dataManager.getPages().size(); i++) {
			if (dataManager.getPages().get(i).getIdentifier().equals(fileName)) {
				page = dataManager.getPages().get(i);
				break;
			}
		}

		return page;
	}

	public void loadPage(String fileName) {
		Page page = identifyPage(fileName);
		loadPage(page);
	}

	// TODO furchtbar
	public void loadPage(Page page) {
		gui.switchToBackgroundPanel();
		currentPage.clean();

		page.initPage(gui.getVerticalResolution());
		// regionManager.update(page.getResized());

		dataManager.setIndex(gui.getPnlPages().getListPages().getSelectedIndex());
		dataManager.setCurrentPage(page);
		setCurrentPage(gui.getDataManager().getCurrentPage());

		gui.getPnlBackground().updateImage(page.getResized(), true);
		gui.pack();
		gui.revalidate();
		gui.repaint();

		SegmentationResult segResult = loadSegResult(currentPage);

		if (segResult != null) {
			currentPage.setSegmentationResult(segResult);
			currentPage.visualizeSegmentation();
			gui.getRbPostProc().setEnabled(true);
			gui.getRbPostProc().doClick();

			return;
		}

		if (currentPage.getSegmentationResult() != null) {
			gui.getRbResult().setEnabled(true);
			gui.getRbPostProc().setEnabled(true);
		}

		if (gui.getRbResult().isSelected() || gui.getRbPostProc().isSelected()) {
			if (currentPage.getResult() != null) {
				gui.getRbResult().setSelected(true);
				gui.getRbPostProc().setSelected(false);
				gui.showImage(currentPage.getResult());
			} else {
				gui.showImage(currentPage.getResized());
				gui.getRbOriginal().setSelected(true);
				gui.getRbBinary().setSelected(false);
				gui.getRbResult().setSelected(false);
				gui.getRbPostProc().setSelected(false);
			}
		} else if (gui.getRbOriginal().isSelected()) {
			gui.showImage(currentPage.getResized());
		} else if (gui.getRbBinary().isSelected()) {
			if (currentPage.getBinary() != null) {
				gui.showImage(currentPage.getBinary());
			} else {
				calcAndShowBinary();
			}
		}

		if (currentPage.getResult() != null) {
			gui.getRbResult().setEnabled(true);
			gui.getRbPostProc().setEnabled(true);
		} else {
			gui.getRbResult().setEnabled(false);
			gui.getRbPostProc().setEnabled(false);
		}

		regionManager.getPointListManager().setPointLists(new ArrayList<PointList>());

		if (gui.getCbSegOnSwitch().isSelected()) {
			segmentCurrentPage();
		}
	}

	public void calcAndShowBinary() {
		int binaryThresh = (int) gui.getPnlProcessParameters().getSpinBinaryThresh().getValue();

		Mat gray = ImageProcessor.calcGray(currentPage.getOriginal());
		Mat binary = new Mat();

		if (binaryThresh == -1) {
			binary = ImageProcessor.calcBinary(gray);
		} else {
			binary = ImageProcessor.calcBinaryFromThresh(gray, binaryThresh);
		}

		currentPage.setBinary(binary);
		gui.showImage(binary);
	}

	public Gui getGui() {
		return gui;
	}

	public void setGui(Gui gui) {
		this.gui = gui;
	}

	public DataManager getDataManager() {
		return dataManager;
	}

	public void setDataManager(DataManager dataManager) {
		this.dataManager = dataManager;
	}

	public Page getCurrentPage() {
		return currentPage;
	}

	public void setCurrentPage(Page currentPage) {
		this.currentPage = currentPage;
	}

	public void startSegmentation() {
		runner = new Thread(this);
		runner.start();
	}

	public void stop() {
		if (runner != null) {
			runner.interrupt();
			runner = null;
		}
	}

	public void run() {
		for (int i = gui.getPnlPages().getListPages().getSelectedIndex(); i < dataManager.getPages().size(); i++) {
			if (keepSegmenting) {
				regionManager.getPointListManager().setPointLists(new ArrayList<PointList>());
				Page page = dataManager.getPages().get(i);

				if (!page.isAccepted()) {
					page.initPage(gui.getVerticalResolution());

					Parameters parameters = new Parameters(regionManager, gui);
					page.setScaleFactor((double) page.getOriginal().height() / parameters.getDesiredImageHeight());

					Segmenter segmenter = new Segmenter(parameters);
					SegmentationResult segmentationResult = segmenter.segment(page.getOriginal());
					page.setSegmentationResult(segmentationResult);

					gui.getPnlPages().getProcessedPages().add(i);
					gui.getPnlPages().getListPages().repaint();

					if (gui.getCbFullAuto().isSelected()) {
						saveCurrentPage(false);
					}
				}
			}
			if (i > 0) {
				dataManager.getPages().get(i - 1).clean();
			}
		}

		setCurrentPage(dataManager.getPages().get(dataManager.getPages().size() - 1));
		currentPage.initPage(gui.getVerticalResolution());
		setKeepSegmenting(false);
		gui.getBtnSegmentAll().setText("Segment All");
	}

	public boolean isKeepSegmenting() {
		return keepSegmenting;
	}

	public String getOutputFolder() {
		return outputFolder;
	}

	public void setOutputFolder(String outputFolder) {
		if (!outputFolder.endsWith(File.separator)) {
			outputFolder += File.separator;
		}

		this.outputFolder = outputFolder;
	}

	public void setKeepSegmenting(boolean keepSegmenting) {
		this.keepSegmenting = keepSegmenting;
	}

	public RegionManager getRegionManager() {
		return regionManager;
	}

	public void setRegionManager(RegionManager regionManager) {
		this.regionManager = regionManager;
	}
}