package gui.drawing;

import java.awt.Cursor;
import java.awt.Point;
import java.awt.Rectangle;

import org.opencv.core.Mat;
import org.opencv.core.Rect;

import positions.Position;

public class PaintHelper {

	private static final int PROX_DIST = 10;

	public static Rectangle resizeRectangle(Point p, Position activePosition, int cursorType, Mat image) {
		Rectangle r = activePosition.getDrawingRect().getBounds();

		int dx = p.x - r.x;
		int dy = p.y - r.y;

		switch (cursorType) {
		case Cursor.N_RESIZE_CURSOR:
			int height = r.height - dy;
			activePosition.updateRect(new Rect(r.x, r.y + dy, r.width, height), image);
			break;
		case Cursor.NW_RESIZE_CURSOR:
			int width = r.width - dx;
			height = r.height - dy;
			activePosition.updateRect(new Rect(r.x + dx, r.y + dy, width, height), image);
			break;
		case Cursor.W_RESIZE_CURSOR:
			width = r.width - dx;
			activePosition.updateRect(new Rect(r.x + dx, r.y, width, r.height), image);
			break;
		case Cursor.SW_RESIZE_CURSOR:
			width = r.width - dx;
			height = dy;
			activePosition.updateRect(new Rect(r.x + dx, r.y, width, height), image);
			break;
		case Cursor.S_RESIZE_CURSOR:
			height = dy;
			activePosition.updateRect(new Rect(r.x, r.y, r.width, height), image);
			break;
		case Cursor.SE_RESIZE_CURSOR:
			width = dx;
			height = dy;
			activePosition.updateRect(new Rect(r.x, r.y, width, height), image);
			break;
		case Cursor.E_RESIZE_CURSOR:
			width = dx;
			activePosition.updateRect(new Rect(r.x, r.y, width, r.height), image);
			break;
		case Cursor.NE_RESIZE_CURSOR:
			width = dx;
			height = r.height - dy;
			activePosition.updateRect(new Rect(r.x, r.y + dy, width, height), image);
			break;
		default:
			System.out.println("unexpected type: " + cursorType);
		}

		return null;
	}

	public static Cursor getCursor(Point p, Rectangle r) {
		int outcode = getOutcode(p, r);

		switch (outcode) {

		case Rectangle.OUT_TOP:
			if (Math.abs(p.y - r.y) < PROX_DIST) {
				return Cursor.getPredefinedCursor(Cursor.N_RESIZE_CURSOR);
			}
			break;
		case Rectangle.OUT_TOP + Rectangle.OUT_LEFT:
			if (Math.abs(p.y - r.y) < PROX_DIST && Math.abs(p.x - r.x) < PROX_DIST) {
				return Cursor.getPredefinedCursor(Cursor.NW_RESIZE_CURSOR);
			}
			break;
		case Rectangle.OUT_LEFT:
			if (Math.abs(p.x - r.x) < PROX_DIST) {
				return Cursor.getPredefinedCursor(Cursor.W_RESIZE_CURSOR);
			}
			break;
		case Rectangle.OUT_LEFT + Rectangle.OUT_BOTTOM:
			if (Math.abs(p.x - r.x) < PROX_DIST && Math.abs(p.y - (r.y + r.height)) < PROX_DIST) {
				return Cursor.getPredefinedCursor(Cursor.SW_RESIZE_CURSOR);
			}
			break;
		case Rectangle.OUT_BOTTOM:
			if (Math.abs(p.y - (r.y + r.height)) < PROX_DIST) {
				return Cursor.getPredefinedCursor(Cursor.S_RESIZE_CURSOR);
			}
			break;
		case Rectangle.OUT_BOTTOM + Rectangle.OUT_RIGHT:
			if (Math.abs(p.x - (r.x + r.width)) < PROX_DIST && Math.abs(p.y - (r.y + r.height)) < PROX_DIST) {
				return Cursor.getPredefinedCursor(Cursor.SE_RESIZE_CURSOR);
			}
			break;
		case Rectangle.OUT_RIGHT:
			if (Math.abs(p.x - (r.x + r.width)) < PROX_DIST) {
				return Cursor.getPredefinedCursor(Cursor.E_RESIZE_CURSOR);
			}
			break;
		case Rectangle.OUT_RIGHT + Rectangle.OUT_TOP:
			if (Math.abs(p.x - (r.x + r.width)) < PROX_DIST && Math.abs(p.y - r.y) < PROX_DIST) {
				return Cursor.getPredefinedCursor(Cursor.NE_RESIZE_CURSOR);
			}
			break;
		default: // center
			return Cursor.getDefaultCursor();
		}

		return null;
	}

	public static boolean checkRect(Rect toCheck, Rectangle area) {
		if (area == null) {
			return false;
		}

		if (area.contains(toCheck.x, toCheck.y, toCheck.width,
				toCheck.height)) {
			return true;
		}

		int new_x = (int) toCheck.x;
		int new_y = (int) toCheck.y;

		if ((toCheck.x + toCheck.width) > area.getWidth()) {
			new_x = (int) area.getWidth() - (int) (toCheck.width - 1);
		}
		if (toCheck.x < 0) {
			new_x = -1;
		}
		if ((toCheck.y + toCheck.height) > area.getHeight()) {
			new_y = (int) area.getHeight() - (int) (toCheck.height - 1);
		}
		if (toCheck.y < 0) {
			new_y = -1;
		}

		toCheck.x = new_x;
		toCheck.y = new_y;

		return false;
	}
	
	private static int getOutcode(Point p, Rectangle rect) {
		Rectangle r = (Rectangle) rect.getBounds().clone();
		r.grow(-PROX_DIST, -PROX_DIST);

		return r.outcode(p.x, p.y);
	}

	public static boolean isOverRect(Point p, Position position) {
		Rectangle r = (Rectangle) position.getDrawingRect().getBounds().clone();
		r.grow(PROX_DIST, PROX_DIST);

		return r.contains(p);
	}
}