package gui.drawing;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.MouseInfo;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.event.MouseEvent;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;

import javax.swing.SwingUtilities;

import org.opencv.core.Rect;

import geometry.PointList;
import geometry.PointListManager;
import positions.Position;
import positions.PositionManager;
import regions.Region;
import regions.RegionManager;

public class Painter {

	private static final int THICKNESS = 2;
	private static final int PERIMETER_MULTIPLIER = 5;

	private BackgroundPanel panel;

	private RegionManager regionManager;
	private PositionManager positionManager;
	private PointListManager pointListManager;

	private Graphics g;
	private Graphics2D g2D;

	private Position activePosition;
	private Rectangle area;

	private PointList linePoints;

	private boolean resizing;
	private boolean drawingRectangle;
	private boolean drawingLines;
	private boolean pressOut;
	private boolean closedLoop;

	private int preX;
	private int preY;

	private Point currentCursorPosition;

	public Painter(BackgroundPanel panel) {
		setRegionManager(panel.getRegionManager());
		setPositionManager(panel.getPositionManager());
		setPointListManager(regionManager.getPointListManager());

		setPanel(panel);

		linePoints = new PointList(new ArrayList<Point>());
	}

	public void handleMouseReleased() {
		pressOut = false;
		resizing = false;

		if (drawingRectangle) {
			drawingRectangle = false;

			new DiaDrawnRect(MouseInfo.getPointerInfo().getLocation(), activePosition, this);
		}
	}

	public void handleMouseDragged(MouseEvent e) {
		if (resizing) {
			if (activePosition != null) {
				PaintHelper.resizeRectangle(e.getPoint(), activePosition, panel.getCursor().getType(),
						panel.getMatImage());
			}
		} else if (drawingRectangle) {
			if (activePosition != null) {
				Rectangle activeRectangle = activePosition.getDrawingRect().getBounds();

				Rect newRect = new Rect((int) activeRectangle.getX(), (int) activeRectangle.getY(),
						(int) (e.getX() - activeRectangle.getX()), (int) (e.getY() - activeRectangle.getY()));
				activePosition.updateRect(newRect, panel.getMatImage());
			}
		} else if (!pressOut) {
			updateLocation(e);
		}

		if (SwingUtilities.isRightMouseButton(e)) {
			if (!drawingRectangle) {
				initDrawingRectangle(e);
			}
		}
	}

	public void handleMouseMoved(MouseEvent e) {
		Point p = e.getPoint();
		currentCursorPosition = p;

		if (drawingLines && linePoints.getPoints().size() > 1) {
			Point hoveredPoint = pointListManager.identifyPoint(p, linePoints, THICKNESS * PERIMETER_MULTIPLIER / 2);

			if (linePoints.getPoints().get(0).equals(hoveredPoint)) {
				setClosedLoop(true);
			} else {
				setClosedLoop(false);
			}
		}

		if (activePosition != null) {
			if (!PaintHelper.isOverRect(p, activePosition)) {
				if (panel.getCursor() != Cursor.getDefaultCursor()) {
					// If cursor is not over rect reset it to the default.
					panel.setCursor(Cursor.getDefaultCursor());
				}
				return;
			} else {
				// Locate cursor relative to center of rect.
				Cursor cursor = PaintHelper.getCursor(p, activePosition.getDrawingRect().getBounds());
				panel.setCursor(cursor);
			}
		}
	}

	public void handleMousePressed(MouseEvent e) {
		if (SwingUtilities.isLeftMouseButton(e)) { // if left click
			if (e.getClickCount() == 2) {
				PointList pointList = pointListManager.identifyPointList(new Point(e.getX(), e.getY()),
						THICKNESS * PERIMETER_MULTIPLIER / 2);

				if (pointList != null) {
					pointList.optimize(panel.getMatImage());
				} else {
					drawingLines = true;
				}
			}

			if (drawingLines) {
				handleLineDrawing(e);
			} else {
				panel.checkAndInitActiveRect(e);
			}

			if (panel.getCursor() != Cursor.getDefaultCursor()) {
				// If cursor is set allow resizing.
				setResizing(true);
			} else { // No resizing but dragging.
				initDraggingRectangle(e);
			}
		} else { // if right click
			if (drawingLines) {
				drawingLines = false;

				if (linePoints.getPoints().size() > 1) {
					pointListManager.getPointLists().add(linePoints);
				}

				setLinePoints(new PointList(new ArrayList<Point>()));
			} else {
				if (!pointListManager.removePointListByPoint(new Point(e.getX(), e.getY()),
						THICKNESS * PERIMETER_MULTIPLIER / 2)
						&& !pointListManager.removePointListByArea(new Point(e.getX(), e.getY()))) {
					// initDrawingRectangle(e);
					Position position = positionManager.identifyActiveRect(e);
					
					if (position != null) {
						positionManager.deletePosition(position);
					}
					
					if(activePosition != null) {
//						area = null;
						setActivePosition(null);
					}
				}
			}
		}
	}

	public void handleLineDrawing(MouseEvent e) {
		if (closedLoop) {
			linePoints.getPoints().add(linePoints.getPoints().get(0));
			setDrawingLines(false);

			PointList newList = new PointList(linePoints.getPoints());
			newList.setClosed(true);
			new PopUpPointList(MouseInfo.getPointerInfo().getLocation(), newList, this);
			pointListManager.getPointLists().add(newList);

			setLinePoints(new PointList(new ArrayList<Point>()));
		} else {
			Point point = new Point(e.getX(), e.getY());
			if (!linePoints.getPoints().contains(point)) {
				linePoints.getPoints().add(point);
			}
		}
	}

	public void initDrawingRectangle(MouseEvent e) {
		if (!drawingRectangle) {
			Rect rect = new Rect(e.getX(), e.getY(), 100, 50);
			Position newPosition = new Position(rect, panel.getMatImage());
			regionManager.addTempPosition(newPosition);
			setActivePosition(newPosition);
		}

		setDrawingRectangle(true);
	}

	// public void initDrawingRectangle(MouseEvent e) {
	// Position position = positionManager.identifyActiveRect(e);
	//
	// if (position != null) {
	// positionManager.deletePosition(position);
	// } else {
	// // TODO
	// Rect rect = new Rect(e.getX(), e.getY(), 100, 50);
	// Position newPosition = new Position(rect, panel.getMatImage());
	// regionManager.addTempPosition(newPosition);
	// setActivePosition(newPosition);
	// }
	//
	// setDrawingRectangle(true);
	// }

	public void initDraggingRectangle(MouseEvent e) {
		if (activePosition != null && activePosition.getDrawingRect().contains(e.getX(), e.getY())) {
			preX = (int) activePosition.getDrawingRect().getX() - e.getX();
			preY = (int) activePosition.getDrawingRect().getY() - e.getY();

			updateLocation(e);
		} else {
			pressOut = true;
		}
	}

	public void updateLocation(MouseEvent e) {
		if (activePosition != null) {
			Rectangle activeRectangle = activePosition.getDrawingRect().getBounds();
			Rect newRect = new Rect(preX + e.getX(), preY + e.getY(), (int) activeRectangle.getWidth(),
					(int) activeRectangle.getHeight());

			PaintHelper.checkRect(newRect, area);
			activePosition.updateRect(newRect, panel.getMatImage());
		}
	}

	public void paint(Graphics g) {
		setG(g);
		setG2D((Graphics2D) g);

		// paints the background image
		g.drawImage(panel.getImage(), 0, 0, null);
		g2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);

		// paints visible regions
		for (Region region : regionManager.getRegions()) {
			if (region.isVisible()) {
				drawRegion(region);
			}
		}

		// drawFixedPositions();

		// paints temporary rectangles
		drawRegion(regionManager.getTempRegion());

		int perimeter = THICKNESS * PERIMETER_MULTIPLIER;
		double radius = perimeter / 2;
		g2D.setStroke(new BasicStroke(THICKNESS));
		g2D.setColor(new Color(128, 0, 255, 255));

		if (linePoints.getPoints().size() > 0) {
			if (!linePoints.isClosed()) {
				drawLines(linePoints, perimeter, radius, true);
			}
		}

		g2D.setColor(new Color(128, 0, 255, 255));

		for (PointList pointList : pointListManager.getPointLists()) {
			if (!pointList.isClosed()) {
				drawLines(pointList, perimeter, radius, false);
			}
		}

		if (closedLoop) {
			ArrayList<Point> points = linePoints.getPoints();
			drawPolygon(points, null);
		}

		drawFixedPointLists();
	}

	// public void drawFixedPositions() {
	// for(Region region : regionManager.getRegions()) {
	// for(Position position : region.getPositions()) {
	// if(position.isFixed()) {
	// if (position.isActive()) {
	// drawRectangle(position.getDrawingRect(), region.getActiveColor());
	// } else {
	// drawRectangle(position.getDrawingRect(), region.getStandardColor());
	// }
	// }
	// }
	// }
	// }

	private void drawPolygon(ArrayList<Point> points, Color color) {
		int[] xPoly = new int[points.size()];
		int[] yPoly = new int[points.size()];

		for (int i = 0; i < points.size(); i++) {
			xPoly[i] = points.get(i).x;
			yPoly[i] = points.get(i).y;
		}

		if (color == null) {
			g2D.setColor(new Color(128, 0, 255, 120));
		} else {
			g2D.setColor(new Color(color.getRed(), color.getGreen(), color.getBlue(), 120));
		}
		Polygon polygon = new Polygon(xPoly, yPoly, points.size());
		g2D.drawPolygon(polygon);
		g2D.fillPolygon(polygon);
	}

	private void drawFixedPointLists() {
		for (PointList pointList : pointListManager.getPointLists()) {
			if (pointList.isClosed()) {
				ArrayList<Point> points = pointList.getPoints();
				drawPolygon(points, RegionManager.getColorByRegionType(pointList.getType()));
			}
		}
	}

	public void drawLines(PointList pointList, int perimeter, double radius, boolean linkToMousePosition) {
		ArrayList<Point> points = pointList.getPoints();

		Point firstPoint = points.get(0);
		Ellipse2D.Double circle = new Ellipse2D.Double(firstPoint.x - radius, firstPoint.y - radius, perimeter,
				perimeter);
		g2D.fill(circle);

		Point lastPoint = firstPoint;

		for (int i = 1; i < points.size(); i++) {
			Point currentPoint = points.get(i);

			g2D.setColor(new Color(128, 0, 255, 70));
			circle = new Ellipse2D.Double(currentPoint.x - radius, currentPoint.y - radius, perimeter, perimeter);
			g2D.fill(circle);
			g2D.setColor(new Color(128, 0, 255, 255));

			g2D.drawLine(lastPoint.x, lastPoint.y, currentPoint.x, currentPoint.y);

			lastPoint = currentPoint;
		}

		if (!drawingLines) {
			g2D.setColor(new Color(128, 0, 255, 255));
			lastPoint = points.get(points.size() - 1);
			circle = new Ellipse2D.Double(lastPoint.x - radius, lastPoint.y - radius, perimeter, perimeter);
			g2D.fill(circle);
		}

		if (drawingLines && linkToMousePosition) {
			g2D.drawLine(lastPoint.x, lastPoint.y, currentCursorPosition.x, currentCursorPosition.y);
		}
	}

	public void drawRegion(Region region) {
		for (Position position : region.getPositions()) {
			if (position.isActive()) {
				drawRectangle(position.getDrawingRect(), region.getActiveColor());
			} else {
				drawRectangle(position.getDrawingRect(), region.getStandardColor());
			}
		}
	}

	public void drawRectangle(Rectangle2D rect, Color color) {
		g2D.setColor(color);
		g2D.draw(rect);

		g2D.setPaint(color);
		g2D.fill(rect);
	}

	public BackgroundPanel getPanel() {
		return panel;
	}

	public void setPanel(BackgroundPanel panel) {
		this.panel = panel;
	}

	public RegionManager getRegionManager() {
		return regionManager;
	}

	public void setRegionManager(RegionManager regionManager) {
		this.regionManager = regionManager;
	}

	public PositionManager getPositionManager() {
		return positionManager;
	}

	public void setPositionManager(PositionManager positionManager) {
		this.positionManager = positionManager;
	}

	public PointListManager getPointListManager() {
		return pointListManager;
	}

	public void setPointListManager(PointListManager pointListManager) {
		this.pointListManager = pointListManager;
	}

	public Graphics getG() {
		return g;
	}

	public void setG(Graphics g) {
		this.g = g;
	}

	public Graphics2D getG2D() {
		return g2D;
	}

	public void setG2D(Graphics2D g2d) {
		g2D = g2d;
	}

	public Rectangle getArea() {
		return area;
	}

	public void setArea(Rectangle area) {
		this.area = area;
	}

	public boolean isDrawingRectangle() {
		return drawingRectangle;
	}

	public void setDrawingRectangle(boolean drawingRectangle) {
		this.drawingRectangle = drawingRectangle;
	}

	public Position getActivePosition() {
		return activePosition;
	}

	public void setActivePosition(Position activePosition) {
		this.activePosition = activePosition;
	}

	public boolean isResizing() {
		return resizing;
	}

	public void setResizing(boolean resizing) {
		this.resizing = resizing;
	}

	public PointList getLinePoints() {
		return linePoints;
	}

	public void setLinePoints(PointList linePoints) {
		this.linePoints = linePoints;
	}

	public boolean isDrawingLines() {
		return drawingLines;
	}

	public void setDrawingLines(boolean drawingLines) {
		this.drawingLines = drawingLines;

		if (!drawingLines) {
			setClosedLoop(false);
		}
	}

	public boolean isClosedLoop() {
		return closedLoop;
	}

	public void setClosedLoop(boolean closedLoop) {
		this.closedLoop = closedLoop;
	}
}