package gui.postProcessing;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.MouseInfo;
import java.awt.Polygon;
import java.awt.RenderingHints;
import java.awt.event.MouseEvent;
import java.awt.geom.Ellipse2D;
import java.awt.image.BufferedImage;

import javax.swing.SwingUtilities;

import org.opencv.core.Mat;
import org.opencv.core.MatOfPoint;
import org.opencv.core.MatOfPoint2f;
import org.opencv.core.Point;
import org.opencv.core.Rect;
import org.opencv.core.Scalar;
import org.opencv.imgproc.Imgproc;

import segmentation.ImageProcessor;
import segmentation.result.ResultRegion;
import segmentation.result.SegmentationResult;

public class PainterPP {

	private Graphics g;

	private ResultsPanel panel;
	private SegmentationResult segmentationResult;

	private BufferedImage image;
	private int desiredImageHeight;

	private MatOfPoint activePosition;
	private Scalar colorActivePosition;

	public PainterPP(ResultsPanel panel, SegmentationResult segmentationResult, int desiredImageHeight) {
		setPanel(panel);
		setSegmentationResult(segmentationResult);
		setDesiredImageHeight(desiredImageHeight);
	}

	public void handleMousePressed(MouseEvent e) {
		Point point = calcPoint(e);

		if (SwingUtilities.isLeftMouseButton(e)) {
			ResultRegion region = markClickedRegion(point);

			if (e.getClickCount() == 2 && region != null) {
				new DiaSwitchType(MouseInfo.getPointerInfo().getLocation(), region, segmentationResult, this);
			}
		} else {
			deleteClickedRegion(point);
		}
	}

	public void deleteClickedRegion(Point point) {
		for (ResultRegion region : segmentationResult.getRegions()) {
			if (pointIsWithinContour(point, region.getPoints())) {
				segmentationResult.removeRegion(region);

				if (activePosition != null && activePosition.equals(region.getPoints())) {
					setActivePosition(null);
					setColorActivePosition(null);
				}
				calcImage();
				return;
			}
		}
	}

	public ResultRegion markClickedRegion(Point point) {
		for (ResultRegion region : segmentationResult.getRegions()) {
			if (pointIsWithinContour(point, region.getPoints())) {
				calcImage();
				setActivePosition(region.getPoints());
				setColorActivePosition(region.getColor());

				return region;
			}
		}

		setActivePosition(null);
		setColorActivePosition(null);
		return null;
	}

	public boolean pointIsWithinContour(Point point, MatOfPoint contour) {
		Rect rect = Imgproc.boundingRect(contour);

		if (rect.contains(point)) {
			if (Imgproc.pointPolygonTest(new MatOfPoint2f(contour.toArray()), point, false) >= 0) {
				return true;
			}
		}

		return false;
	}

	public void calcImage() {
		Mat resized = ImageProcessor.resize(panel.getOriginal(), desiredImageHeight);
		Mat result = segmentationResult.drawResult(panel.getOriginal(), resized);
		BufferedImage image = ImageProcessor.matToBufferedImage(result);

		panel.setPreferredSize(new Dimension(image.getWidth(null), image.getHeight(null)));
		panel.setSize(new Dimension(image.getWidth(null), image.getHeight(null)));

		setImage(image);
	}

	public void paint(Mat original) {
		Graphics2D g2D = (Graphics2D) g;

		// paints the background image
		g.drawImage(image, 0, 0, null);
		g2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);

		// paints the active region
		if (activePosition != null && colorActivePosition != null) {
			double scaleFactor = (double) segmentationResult.getVerticalResolution() / desiredImageHeight;
			Point[] points = activePosition.toArray();
			int[] xPoly = new int[points.length];
			int[] yPoly = new int[points.length];

			for (int i = 0; i < points.length; i++) {
				xPoly[i] = (int) (points[i].x / scaleFactor);
				yPoly[i] = (int) (points[i].y / scaleFactor);
			}

			Polygon poly = new Polygon(xPoly, yPoly, xPoly.length);

			Color color = new Color((int) colorActivePosition.val[2], (int) colorActivePosition.val[1],
					(int) colorActivePosition.val[0], 150);
			g.setColor(color);
			g.drawPolygon(poly);
			g.fillPolygon(poly);
		}

		g.setColor(new Color(255, 0, 128, 128));
		Graphics2D g2 = (Graphics2D) g;
		g2.setStroke(new BasicStroke(5));

		if (segmentationResult.getReadingOrder().size() > 0) {
			ResultRegion lastRegion = segmentationResult.getReadingOrder().get(0);

			for (int i = 1; i < segmentationResult.getReadingOrder().size(); i++) {
				ResultRegion currectRegion = segmentationResult.getReadingOrder().get(i);

				if (currectRegion != null) {
					Ellipse2D.Double circle = new Ellipse2D.Double(lastRegion.getScaledCenterOfGravity().getX() - 10,
							(int) lastRegion.getScaledCenterOfGravity().getY() - 10, 20, 20);
					g2D.fill(circle);

					g.drawLine((int) lastRegion.getScaledCenterOfGravity().getX(),
							(int) lastRegion.getScaledCenterOfGravity().getY(),
							(int) currectRegion.getScaledCenterOfGravity().getX(),
							(int) currectRegion.getScaledCenterOfGravity().getY());
					lastRegion = currectRegion;
				}
			}
		}
	}

	public Point calcPoint(MouseEvent e) {
		double scaleFactor = (double) segmentationResult.getVerticalResolution() / desiredImageHeight;
		Point point = new Point(scaleFactor * e.getX(), scaleFactor * e.getY());

		return point;
	}

	public Graphics getG() {
		return g;
	}

	public void setG(Graphics g) {
		this.g = g;
	}

	public ResultsPanel getPanel() {
		return panel;
	}

	public void setPanel(ResultsPanel panel) {
		this.panel = panel;
	}

	public SegmentationResult getSegmentationResult() {
		return segmentationResult;
	}

	public void setSegmentationResult(SegmentationResult segmentationResult) {
		this.segmentationResult = segmentationResult;
	}

	public BufferedImage getImage() {
		return image;
	}

	public void setImage(BufferedImage image) {
		this.image = image;
	}

	public int getDesiredImageHeight() {
		return desiredImageHeight;
	}

	public void setDesiredImageHeight(int desiredImageHeight) {
		this.desiredImageHeight = desiredImageHeight;
	}

	public MatOfPoint getActivePosition() {
		return activePosition;
	}

	public void setActivePosition(MatOfPoint activePosition) {
		this.activePosition = activePosition;
	}

	public Scalar getColorActivePosition() {
		return colorActivePosition;
	}

	public void setColorActivePosition(Scalar colorActivePosition) {
		this.colorActivePosition = colorActivePosition;
	}
}