package gui.readingOrder;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Polygon;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.image.BufferedImage;
import java.util.ArrayList;

import javax.swing.JPanel;
import javax.swing.SwingUtilities;

import org.opencv.core.Mat;
import org.opencv.core.Point;
import dataManagement.Page;
import regions.type.RegionType;
import segmentation.ImageProcessor;
import segmentation.result.ResultRegion;
import segmentation.result.SegmentationResult;

@SuppressWarnings("serial")
public class ReadingOrderPainter extends JPanel implements KeyListener {

	private Image image;
	private Page page;
	private SegmentationResult segmentationResult;

	private ResultRegion activeResult;
	private boolean drawingReadingOrder;
	private boolean assigningRO;
	private Point activePoint;

	public ReadingOrderPainter(Page page) {
		setSegmentationResult(page.getSegmentationResult());
		setPage(page);

		updateImage(page.getResized());
		setOpaque(false);

		addListeners();
		
		setFocusable(true);
        requestFocusInWindow();
		addKeyListener(this);
	}

	public void addListeners() {
		addMouseListener(new MouseAdapter() {
			public void mousePressed(MouseEvent e) {
				if (assigningRO) {
					if (SwingUtilities.isLeftMouseButton(e)) {
						Point point = new Point(e.getX(), e.getY());
						setActivePoint(point);
						ResultRegion result = segmentationResult.identifyResult(point);

						if (result != null && !result.getType().equals(RegionType.image) && !result.isAssigned()) {
							result.setAssigned(true);
							result.setReadingOrderIndex(segmentationResult.getReadingOrder().size());
							segmentationResult.getReadingOrder().add(result);
							repaint();
						}
					} else if (SwingUtilities.isRightMouseButton(e)) {
						setAssigningRO(false);
						setActivePoint(null);
					}
				}
			}
		});

		addMouseMotionListener(new MouseMotionAdapter() {
			public void mouseMoved(MouseEvent e) {
				Point point = new Point(e.getX(), e.getY());
				setActivePoint(point);
				ResultRegion result = segmentationResult.identifyResult(point);

				if (result != null) {
					if (assigningRO) {
						if (!result.isAssigned()) {
							setActiveResult(result);

						}
					} else {
						setActiveResult(result);
					}
					// result.setActive(true);
				} else {
					setActiveResult(null);
				}

				repaint();
			}
		});
	}

	public void paint(Graphics g) {
		super.paint(g);

		// paint image
		Mat result = page.getOriginal();
		result = segmentationResult.drawResult(result, page.getResized());

		setImage(ImageProcessor.matToBufferedImage(result));
		g.drawImage(image, 0, 0, null);

		// paint the active region
		if (activeResult != null) {
			Polygon poly = new Polygon(activeResult.getScaledPolygon().getX_coords(),
					activeResult.getScaledPolygon().getY_coords(), activeResult.getScaledPolygon().getN());

			Color color = new Color((int) activeResult.getColor().val[2], (int) activeResult.getColor().val[1],
					(int) activeResult.getColor().val[0], 150);
			g.setColor(color);
			// g.drawPolygon(poly);
			g.fillPolygon(poly);
		}

		if (assigningRO) {
			// fill already assigned regions
			Color color = new Color(130, 130, 130, 150);
			g.setColor(color);

			for (ResultRegion region : segmentationResult.getRegions()) {
				if (region.isAssigned()) {
					Polygon poly = new Polygon(region.getScaledPolygon().getX_coords(),
							region.getScaledPolygon().getY_coords(), region.getScaledPolygon().getN());

					g.fillPolygon(poly);
				}
			}
		}

		// show reading order
		g.setColor(new Color(255, 0, 128));
		Graphics2D g2 = (Graphics2D) g;
		g2.setStroke(new BasicStroke(5));

		if (segmentationResult.getReadingOrder().size() > 0) {
			ResultRegion lastRegion = segmentationResult.getReadingOrder().get(0);

			for (int i = 1; i < segmentationResult.getReadingOrder().size(); i++) {
				ResultRegion currectRegion = segmentationResult.getReadingOrder().get(i);

				if (currectRegion != null) {
					g.drawLine((int) lastRegion.getScaledCenterOfGravity().getX(),
							(int) lastRegion.getScaledCenterOfGravity().getY(),
							(int) currectRegion.getScaledCenterOfGravity().getX(),
							(int) currectRegion.getScaledCenterOfGravity().getY());
					lastRegion = currectRegion;
				}
			}

			if (assigningRO && activePoint != null) {
				g.drawLine((int) lastRegion.getScaledCenterOfGravity().getX(),
						(int) lastRegion.getScaledCenterOfGravity().getY(), (int) activePoint.x, (int) activePoint.y);
			}
		}
	}

	public void initNewRO() {
		segmentationResult.setReadingOrder(new ArrayList<ResultRegion>());

		for (ResultRegion region : segmentationResult.getRegions()) {
			region.setAssigned(false);
			region.setReadingOrderIndex(-1);
		}

		setAssigningRO(true);
	}

	public void updateImage(Mat image) {
		BufferedImage inputBI = ImageProcessor.matToBufferedImage(image);
		setPreferredSize(new Dimension(inputBI.getWidth(null), inputBI.getHeight(null)));
		setSize(new Dimension(inputBI.getWidth(null), inputBI.getHeight(null)));
		this.image = inputBI;

		segmentationResult.calcROBinaries(image);
		
		repaint();
	}

	public Image getImage() {
		return image;
	}

	public void setImage(Image image) {
		this.image = image;
	}

	public Page getPage() {
		return page;
	}

	public void setPage(Page page) {
		this.page = page;
	}

	public SegmentationResult getSegmentationResult() {
		return segmentationResult;
	}

	public void setSegmentationResult(SegmentationResult segmentationResult) {
		this.segmentationResult = segmentationResult;
	}

	public ResultRegion getActiveResult() {
		return activeResult;
	}

	public void setActiveResult(ResultRegion activeResult) {
		this.activeResult = activeResult;
	}

	public boolean isDrawingReadingOrder() {
		return drawingReadingOrder;
	}

	public void setDrawingReadingOrder(boolean drawingReadingOrder) {
		this.drawingReadingOrder = drawingReadingOrder;
	}

	public boolean isAssigningRO() {
		return assigningRO;
	}

	public void setAssigningRO(boolean assigningRO) {
		this.assigningRO = assigningRO;
	}

	public Point getActivePoint() {
		return activePoint;
	}

	public void setActivePoint(Point activePoint) {
		this.activePoint = activePoint;
	}

	@Override
	public void keyPressed(KeyEvent e) {
		if (e.getKeyCode() == KeyEvent.VK_N) {
            System.out.println("Right key typed");
        }
	}

	@Override
	public void keyReleased(KeyEvent e) {
	}

	@Override
	public void keyTyped(KeyEvent e) {
	}
}