package linePP.auto;

import java.util.ArrayList;
import org.opencv.core.Core;
import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfPoint;
import org.opencv.core.Point;
import org.opencv.core.Rect;
import org.opencv.core.Scalar;
import org.opencv.imgproc.Imgproc;

import geometry.Ray;
import geometry.StraightLine;
import imageProcessing.Contour;
import imageProcessing.ImageProcessor;
import linePP.data.Line;
import linePP.data.LinePage;
import linePP.data.Segment;
import regions.type.RegionType;

public class Logic {

	private static final double REQUIRED_HEIGHT_FACTOR = 1.3;
	private static final int INITIAL_INDENT = 150;
	private static final int HEADING_INDENT = 30;
	private static final int MAX_LINE_BACKTRACK = 3;
	private static final int BIG_LETTER_THRESH = 1500;
	private static final int SMALL_LETTER_THRESH = 200;

	public static void doMarkus(LinePage linePage) {
		drawBinary(linePage);

		assignSide(linePage);
		calcLeftLines(linePage);

		double avgLineHeight = calcAverageLineHeight(linePage);
		detectHighLines(linePage, avgLineHeight);
		pruneHighLines(linePage);
		findLinesBelowHeading(linePage);
		findLinesAboveHeading(linePage);
	}

	public static void findLinesAboveHeading(LinePage linePage) {
		for (Segment segment : linePage.getSegments()) {
			if (!segment.getType().equals(RegionType.heading)) {
				for (int i = 0; i < segment.getLines().size(); i++) {
					if (segment.getLines().get(i).getLineType().equals(LineType.HIGH)) {
						boolean run = true;
						int startIndex = i - 1;
						int cnt = 0;
						int lowestHeadingIndex = -1;

						while (run) {
							if (startIndex > -1) {
								Line toCheck = segment.getLines().get(startIndex);

								if (isIndented(linePage, toCheck, segment, HEADING_INDENT)) {
									toCheck.setLineType(LineType.HEADING);
									lowestHeadingIndex = startIndex;
									startIndex--;
								} else {
									startIndex--;
								}
							} else {
								run = false;
							}

							cnt++;

							if (cnt >= MAX_LINE_BACKTRACK) {
								run = false;
							}
						}

						if (lowestHeadingIndex > -1) {
							for (int k = lowestHeadingIndex; k < i; k++) {
								segment.getLines().get(k).setLineType((LineType.HEADING));
							}
						}
					}
				}
			}
		}
	}

	private static Point detectLeft(LinePage linePage, Segment segment, Line toCheck) {
		if (segment.getLeftLine() == null) {
			segment.calcLeftLine(linePage.getOriginal());
		}

		StraightLine straightLine = segment.getLeftLine();

		Rect rect = Imgproc.boundingRect(toCheck.getLinePoints());
		Mat mask = new Mat(linePage.getOriginal().size(), CvType.CV_8UC1, new Scalar(0));

		// extend lineRect to straightLine
		// {tl, tr, br, bl}
		Point[] linePoints = ImageProcessor.findLinePointOrder(toCheck.getLinePoints());
		Ray topRay = new Ray(linePoints[0], linePoints[1]);
		Ray bottomRay = new Ray(linePoints[2], linePoints[3]);
		Ray leftRay = new Ray(straightLine.getTl(), straightLine.getBl());
		Point intTop = topRay.calcIntersectionPoint(leftRay);
		Point intBottom = bottomRay.calcIntersectionPoint(leftRay);

		// Ray ray1 = new Ray(new Point(0.5,2), 1, 1);
		// Ray ray2 = new Ray(new Point(-0.5,7), -0.5, 1);
		// System.out.println(ray1.calcIntersectionPoint(ray2));

		Point[] tempPoints = new Point[4];
		tempPoints[0] = intTop;
		tempPoints[1] = linePoints[1];
		tempPoints[2] = linePoints[2];
		tempPoints[3] = intBottom;
		MatOfPoint points = new MatOfPoint(tempPoints);
		rect = Imgproc.boundingRect(points);

		Core.fillConvexPoly(mask, new MatOfPoint(points), new Scalar(255));

		// Mat temp = linePage.getOriginal().clone();
		// Core.fillConvexPoly(temp, new MatOfPoint(points), new Scalar(255));
		// Core.line(temp, segment.getLeftLine().getTl(),
		// segment.getLeftLine().getBl(), new Scalar(0, 255, 0), 5);
		// Core.circle(temp, intTop, 5, new Scalar(0, 0, 255), -1);
		// Core.circle(temp, intBottom, 5, new Scalar(0, 0, 255), -1);
		// Highgui.imwrite("C:/Users/Reul/Desktop/tempResult.png", temp);

		Mat tempImage = new Mat(linePage.getOriginal().size(), linePage.getOriginal().type(),
				new Scalar(255, 255, 255));
		Mat gray = new Mat();
		Mat binary = new Mat();

		linePage.getOriginal().copyTo(tempImage, mask);
		Mat image = new Mat(tempImage, rect);

		Imgproc.cvtColor(image, gray, Imgproc.COLOR_BGR2GRAY);
		Imgproc.threshold(gray, binary, -1, 255, Imgproc.THRESH_OTSU);
		Core.bitwise_not(binary, binary);

		ArrayList<MatOfPoint> contours = Contour.findContours(binary);

		int left = Integer.MAX_VALUE;
		Rect leftRect = null;

		ArrayList<MatOfPoint> acceptedContours = new ArrayList<MatOfPoint>();
		for (int i = 0; i < contours.size(); i++) {
			MatOfPoint contour = contours.get(i);
			double area = Imgproc.contourArea(contour);
			Rect tempRect = Imgproc.boundingRect(contour);

			if (area > SMALL_LETTER_THRESH) {
				if (tempRect.contains(new Point((tempRect.x + tempRect.br().x) / 2, binary.height() / 3))
						|| tempRect.contains(new Point((tempRect.x + tempRect.br().x) / 2, 2 * binary.height() / 3))) {
					acceptedContours.add(contour);
					if (tempRect.x < left) {
						// ImageProcessor.showResult(binary, binary.size());
						left = tempRect.x;
						leftRect = tempRect;
					}
				} else {
				}
			}

		}

		// Imgproc.drawContours(image, acceptedContours, -1, new Scalar(0, 0,
		// 255));
		// ImageProcessor.showResult(image, image.size());

		Point resultPoint = new Point(rect.x + leftRect.x, rect.y + (leftRect.tl().y + leftRect.br().y) / 2);

		return resultPoint;
	}

	private static boolean isIndented(LinePage linePage, Line toCheck, Segment segment, int thresh) {
		Point point = detectLeft(linePage, segment, toCheck);

		if (segment.getLeftLine() == null) {
			segment.calcLeftLine(linePage.getOriginal());
		}

		StraightLine straightLine = segment.getLeftLine();

		if (!straightLine.isSlopIsInf()) {
			double x = straightLine.calcX(point.y);
			double diff = Math.abs(point.x - x);

			if (diff > thresh) {
				return true;
			}
		}

		return false;
	}

	public static void findLinesBelowHeading(LinePage linePage) {
		for (Segment segment : linePage.getSegments()) {
			if (!segment.getType().equals(RegionType.heading)) {
				for (int i = 0; i < segment.getLines().size(); i++) {
					if (segment.getLines().get(i).getLineType().equals(LineType.HIGH)
							&& isIndented(linePage, segment.getLines().get(i), segment, INITIAL_INDENT)) {
						Line firstLine = null;
						Line secondLine = null;

						if (i + 1 < segment.getLines().size()) {
							firstLine = segment.getLines().get(i + 1);

							if (i + 2 < segment.getLines().size()) {
								secondLine = segment.getLines().get(i + 2);
							}
						}

						if (firstLine != null) {
							firstLine.setLineType(LineType.BELOW_HIGH);
							if (secondLine != null && !firstLine.getLineType().equals(LineType.HIGH)) {
								secondLine.setLineType(LineType.BELOW_HIGH);
							}
						}
					}
				}
			}
		}
	}

	private static boolean isTrueHighLine(Line toCheck, LinePage linePage) {
		Rect rect = Imgproc.boundingRect(toCheck.getLinePoints());
		Mat mask = new Mat(linePage.getOriginal().size(), CvType.CV_8UC1, new Scalar(0));

		Core.fillConvexPoly(mask, toCheck.getLinePoints(), new Scalar(255));

		Mat tempImage = new Mat(linePage.getOriginal().size(), linePage.getOriginal().type(),
				new Scalar(255, 255, 255));
		Mat gray = new Mat();
		Mat binary = new Mat();

		linePage.getOriginal().copyTo(tempImage, mask);
		Mat image = new Mat(tempImage, rect);
		// Highgui.imwrite("C:/Users/Reul/Desktop/tempResult.png", image);

		Imgproc.cvtColor(image, gray, Imgproc.COLOR_BGR2GRAY);
		Imgproc.threshold(gray, binary, -1, 255, Imgproc.THRESH_OTSU);
		Core.bitwise_not(binary, binary);

		// ImageProcessor.showResult(image, image.size());

		ArrayList<MatOfPoint> contours = Contour.findContours(binary);

		double areaSum = 0;
		double avgArea = 0;
		int cnt = 0;

		for (MatOfPoint contour : contours) {
			double area = Imgproc.contourArea(contour);

			if (area > 100) {
				areaSum += area;
				cnt++;
			}
		}

		if (cnt > 0) {
			avgArea = areaSum / cnt;
		}

		Mat result = new Mat(image.size(), image.type(), new Scalar(255, 255, 255));
		Imgproc.drawContours(result, contours, -1, new Scalar(0, 0, 255));
		Imgproc.drawContours(result, contours, 0, new Scalar(255, 0, 0));
		// ImageProcessor.showResult(result, result.size());

		if (avgArea > BIG_LETTER_THRESH) {
			return true;
		}

		return false;
	}

	private static void pruneHighLines(LinePage linePage) {
		Mat image = linePage.getOriginal().clone();

		for (Segment segment : linePage.getSegments()) {
			if (!segment.getType().equals(RegionType.heading)) {
				for (int i = 0; i < segment.getLines().size(); i++) {
					Line toCheck = segment.getLines().get(i);

					if (toCheck.getLineType().equals(LineType.HIGH_CANDIDATE)) {
						if (isTrueHighLine(toCheck, linePage)) {
							segment.getLines().get(i).setLineType(LineType.HIGH);

							if (segment.getLeftLine() == null) {
								segment.calcLeftLine(linePage.getOriginal().clone());
							}
						}
					}
				}
			}

			if (segment.getLeftLine() != null) {
				Core.line(image, segment.getLeftLine().getTl(), segment.getLeftLine().getBl(), new Scalar(0, 255, 0),
						5);
			}
		}
		// Highgui.imwrite("C:/Users/Reul/Desktop/tempResult.png", image);
	}

	private static void detectHighLines(LinePage linePage, double avgLineHeight) {
		for (Segment segment : linePage.getSegments()) {
			if (!segment.getType().equals(RegionType.heading)) {
				for (int i = 0; i < segment.getLines().size(); i++) {
					double height = segment.getLines().get(i).getBottomLeft().y
							- segment.getLines().get(i).getTopLeft().y;
					double width = segment.getLines().get(i).getBottomRight().x
							- segment.getLines().get(i).getTopLeft().x;

					// TODO
					if (height > REQUIRED_HEIGHT_FACTOR * avgLineHeight && width > 100) {
						segment.getLines().get(i).setLineType(LineType.HIGH_CANDIDATE);
					}
				}
			}
		}
	}

	private static double calcAverageLineHeight(LinePage linePage) {
		double sumHeight = 0;
		int cnt = 0;

		for (Segment segment : linePage.getSegments()) {
			if (!segment.getType().equals(RegionType.heading)) {
				for (Line line : segment.getLines()) {
					sumHeight += line.getBottomLeft().y - line.getTopLeft().y;
					cnt++;
				}
			}
		}

		double avgLineHeight = sumHeight / cnt;
		// System.out.println("avgLineHeight: " + avgLineHeight + "\n");

		return avgLineHeight;
	}

	private static StraightLine calcLine(ArrayList<Line> lines) {
		ArrayList<Point> pointList = new ArrayList<Point>();

		double top = Double.MAX_VALUE;
		double bottom = 0;

		for (Line line : lines) {
			Point point = line.getTopLeft();
			pointList.add(point);

			if (point.y < top) {
				top = point.y;
			}

			if (point.y > bottom) {
				bottom = point.y;
			}
		}

		Point[] points = pointList.toArray(new Point[pointList.size()]);
		Mat fittedLine = new Mat();
		Imgproc.fitLine(new MatOfPoint(points), fittedLine, Imgproc.CV_DIST_L1, 0, 0.01, 0.01);

		double vw = fittedLine.get(0, 0)[0];
		double vy = fittedLine.get(1, 0)[0];
		double x0 = fittedLine.get(2, 0)[0];
		double y0 = fittedLine.get(3, 0)[0];

		StraightLine straightLine = new StraightLine(x0, y0, vw, vy, top, bottom);

		return straightLine;
	}

	private static void calcLeftLines(LinePage linePage) {
		ArrayList<Line> leftLines = new ArrayList<Line>();
		ArrayList<Line> rightLines = new ArrayList<Line>();

		for (Segment segment : linePage.getSegments()) {
			if (!segment.getType().equals(RegionType.heading)) {
				if (segment.isLeft()) {
					leftLines.addAll(segment.getLines());
				} else {
					rightLines.addAll(segment.getLines());
				}
			}
		}

		StraightLine leftStraightLine = calcLine(leftLines);
		StraightLine rightStraightLine = calcLine(rightLines);

		Mat image = linePage.getOriginal().clone();
		Core.line(image, leftStraightLine.getTl(), leftStraightLine.getBl(), new Scalar(0, 255, 0), 5);
		Core.line(image, rightStraightLine.getTl(), rightStraightLine.getBl(), new Scalar(0, 255, 0), 5);
//		Highgui.imwrite("C:/Users/Reul/Desktop/tempResult.png", image);

		for (Segment segment : linePage.getSegments()) {
			if (!segment.getType().equals(RegionType.heading)) {
				if (segment.isLeft()) {
					segment.setLeftLine(leftStraightLine);
				} else {
					segment.setLeftLine(rightStraightLine);
				}
			}
		}
	}

	private static void assignSide(LinePage linePage) {
		int border = linePage.getOriginal().width() / 2;

		for (Segment segment : linePage.getSegments()) {
			Rect rect = Imgproc.boundingRect(segment.getRegion());

			if (rect.tl().x + rect.width / 2 < border) {
				segment.setLeft(true);
			}
		}
	}

	public static void drawBinary(LinePage linePage) {
		ArrayList<MatOfPoint> contours = new ArrayList<MatOfPoint>();

		for (Segment segment : linePage.getSegments()) {
			if (!segment.getType().equals(RegionType.heading)) {
				contours.add(segment.getRegion());
			}
		}

		// Mat result = linePage.getOriginal().clone();
		// Imgproc.drawContours(result, contours, -1, new Scalar(0, 0, 255),
		// -1);
		// Highgui.imwrite("C:/Users/Reul/Desktop/tempResult.png", result);
	}
}