package linePP.auto;

import java.util.ArrayList;
import org.opencv.core.Core;
import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfPoint;
import org.opencv.core.Rect;
import org.opencv.core.Scalar;
import org.opencv.imgproc.Imgproc;

import imageProcessing.Contour;
import linePP.data.Line;
import linePP.data.LinePage;
import linePP.data.Segment;
import regions.type.RegionType;

public class Logic954 {

	private static final double LEFT_RIGHT_BORDER_PERCENTAGE = 0.6;
	private static final double REQUIRED_HEIGHT_FACTOR = 1.0;
	private static final int BIG_LETTER_AVERAGE_THRESH = 2500;
	private static final int BIG_LETTER_CANDIDATE_THRESH = 1000;

	public static void doMarkus(LinePage linePage) {
		drawBinary(linePage);
		pruneSegments(linePage);
		assignSide(linePage);

		double avgLineHeight = calcAverageLineHeight(linePage);
		detectHighLines(linePage, avgLineHeight);
		pruneHighLines(linePage);
	}

	private static boolean isTrueHighLine(Line toCheck, LinePage linePage) {
		Rect rect = Imgproc.boundingRect(toCheck.getLinePoints());
		Mat mask = new Mat(linePage.getOriginal().size(), CvType.CV_8UC1, new Scalar(0));

		Core.fillConvexPoly(mask, toCheck.getLinePoints(), new Scalar(255));

		Mat tempImage = new Mat(linePage.getOriginal().size(), linePage.getOriginal().type(),
				new Scalar(255, 255, 255));
		Mat gray = new Mat();
		Mat binary = new Mat();

		linePage.getOriginal().copyTo(tempImage, mask);
		Mat image = new Mat(tempImage, rect);

		Imgproc.cvtColor(image, gray, Imgproc.COLOR_BGR2GRAY);
		Imgproc.threshold(gray, binary, -1, 255, Imgproc.THRESH_OTSU);
		Core.bitwise_not(binary, binary);

		ArrayList<MatOfPoint> contours = Contour.findContours(binary);

		double areaSum = 0;
		double avgArea = 0;
		int cnt = 0;

		for (MatOfPoint contour : contours) {
			double area = Imgproc.contourArea(contour);

			if (area > BIG_LETTER_CANDIDATE_THRESH) {
				areaSum += area;
				cnt++;
			}
		}

		if (cnt > 0) {
			avgArea = areaSum / cnt;
		}

		Mat result = new Mat(image.size(), image.type(), new Scalar(255, 255, 255));
		Imgproc.drawContours(result, contours, -1, new Scalar(0, 0, 255));
		Imgproc.drawContours(result, contours, 0, new Scalar(255, 0, 0));

		if (avgArea > BIG_LETTER_AVERAGE_THRESH) {
			return true;
		}

		return false;
	}

	private static void pruneHighLines(LinePage linePage) {
		for (Segment segment : linePage.getSegments()) {
			if (!segment.getType().equals(RegionType.heading)) {
				for (int i = 0; i < segment.getLines().size(); i++) {
					Line toCheck = segment.getLines().get(i);

					if (toCheck.getLineType().equals(LineType.HIGH_CANDIDATE)) {
						if (isTrueHighLine(toCheck, linePage)) {
							segment.getLines().get(i).setLineType(LineType.HIGH);
						}
					}
				}
			}
		}
	}

	private static void detectHighLines(LinePage linePage, double avgLineHeight) {
		for (Segment segment : linePage.getSegments()) {
			if (!segment.getType().equals(RegionType.heading)) {
				for (int i = 0; i < segment.getLines().size(); i++) {
					double height = segment.getLines().get(i).getBottomLeft().y
							- segment.getLines().get(i).getTopLeft().y;
					double width = segment.getLines().get(i).getBottomRight().x
							- segment.getLines().get(i).getTopLeft().x;

					if (height > REQUIRED_HEIGHT_FACTOR * avgLineHeight && width > 100) {
						segment.getLines().get(i).setLineType(LineType.HIGH_CANDIDATE);
					}
				}
			}
		}
	}

	private static double calcAverageLineHeight(LinePage linePage) {
		double sumHeight = 0;
		int cnt = 0;

		for (Segment segment : linePage.getSegments()) {
			if (!segment.getType().equals(RegionType.heading)) {
				for (Line line : segment.getLines()) {
					sumHeight += line.getBottomLeft().y - line.getTopLeft().y;
					cnt++;
				}
			}
		}

		double avgLineHeight = sumHeight / cnt;

		return avgLineHeight;
	}

	private static void assignSide(LinePage linePage) {
		int border = (int) (LEFT_RIGHT_BORDER_PERCENTAGE * linePage.getOriginal().width());

		for (Segment segment : linePage.getSegments()) {
			Rect rect = Imgproc.boundingRect(segment.getRegion());

			if (rect.tl().x + rect.width / 2 < border) {
				segment.setLeft(true);
			}
		}
	}

	private static void pruneSegments(LinePage linePage) {
		ArrayList<Segment> toRemove = new ArrayList<Segment>();

		for (Segment segment : linePage.getSegments()) {
			if (!segment.getType().equals(RegionType.heading) && !segment.getType().equals(RegionType.page_number)
					&& !segment.getType().equals(RegionType.image) && segment.getLines().size() == 0) {
				toRemove.add(segment);
			}
		}

		linePage.getSegments().removeAll(toRemove);
	}

	// TODO ???
	public static void drawBinary(LinePage linePage) {
		ArrayList<MatOfPoint> contours = new ArrayList<MatOfPoint>();

		for (Segment segment : linePage.getSegments()) {
			if (!segment.getType().equals(RegionType.heading)) {
				contours.add(segment.getRegion());
			}
		}
	}
}