package linePP.calc;

import java.awt.image.BufferedImage;
import java.nio.ByteBuffer;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.ArrayList;

import org.opencv.core.Mat;
import org.opencv.core.MatOfPoint;
import org.opencv.core.Point;
import org.opencv.core.Scalar;
import org.opencv.imgproc.Imgproc;

import imageProcessing.ImageProcessor;
import net.sourceforge.tess4j.ITessAPI.TessBaseAPI;
import net.sourceforge.tess4j.ITessAPI.TessPageIterator;
import net.sourceforge.tess4j.ITessAPI.TessPageSegMode;
import net.sourceforge.tess4j.TessAPI;
import net.sourceforge.tess4j.TessAPI1;
import net.sourceforge.tess4j.util.ImageIOHelper;
import net.sourceforge.tess4j.util.LoadLibs;

public class LineSeparator {

	public static int cnt = 0;

	private Page page;
	private Mat image;
	private Scalar avgBgd;

	public LineSeparator(Page page, Mat image, boolean useAvgBgd) {
		setPage(page);
		setImage(image);

		double[] avgBgd = ImageProcessor.calcAverageBackground(image);

		if (useAvgBgd) {
			if (avgBgd.length > 1) {
				setAvgBgd(new Scalar(avgBgd[2], avgBgd[1], avgBgd[0]));
			} else {
				setAvgBgd(new Scalar(avgBgd[0]));
			}
		} else {
			setAvgBgd(new Scalar(255, 255, 255));
		}
	}

	public void calcAndSetWeightedAngle(ArrayList<Segment> segments) {
		double weightedAngleSum = 0;
		double areaSum = 0;

		for (Segment segment : segments) {
			if (segment.isBigEnough()) {
				double area = Imgproc.contourArea(segment.getPolygon());
				areaSum += area;
				weightedAngleSum += area * segment.getAngle();
			}
		}

		if (areaSum != 0) {
			double weightedAngle = weightedAngleSum / areaSum;
			System.out.println("	Weighted angle: " + weightedAngle);

			for (Segment segment : segments) {
				if (segment.isBigEnough()) {
					segment.setAngle(weightedAngle);
				}
			}
		}
	}

	public ArrayList<Segment> calcLines(boolean postProcess) {
		ArrayList<Segment> lineSegments = new ArrayList<Segment>();

		System.out.println("Processing " + page.getSegments().size() + " Segments");
		System.out.println("Calculating Angles:");
		int cnt = 0;

		for (Segment segment : page.getSegments()) {
			System.out.println("	Processing Segment " + cnt);

			if (!segment.getType().equals("image")) {
				Mat segmentImage = extractSegment(segment.getPolygon());
				segment.setSegmentImage(segmentImage);

				if (segment.getSegmentImage() == null) {
					System.out.println("null");
				}

				float angle = determineAngle(segmentImage);

				if (angle == -360.0) {
					segment.setBigEnough(false);
				} else {
					segment.setBigEnough(true);
					segment.setAngle(angle);
				}
			} else {
				System.out.println("	Skipping Image!");
			}

			cnt++;
		}

		calcAndSetWeightedAngle(page.getSegments());
		System.gc();

		System.out.println("Segmenting Lines:");
		cnt = 0;

		for (Segment segment : page.getSegments()) {
			System.out.println("	Processing Segment " + cnt);

			if (segment.getSegmentImage() != null) {
				ArrayList<MatOfPoint> lines = findLines(segment.getSegmentImage(), segment, postProcess);

				if (lines != null) {
					for (MatOfPoint line : lines) {
						Segment newSegment = new Segment(segment.getType(), line);
						lineSegments.add(newSegment);
					}
				}
			}

			cnt++;
		}

		System.out.println("Done!");

		return lineSegments;
	}

	public ArrayList<MatOfPoint> findLines(Mat segmentImage, Segment segment, boolean postProcess) {
		ArrayList<MatOfPoint> lines = new ArrayList<MatOfPoint>();

		double angle = segment.getAngle();
		double degAngle = angle * 180 / Math.PI;

		// Highgui.imwrite("C:/Users/Reul/Desktop/output.png", segmentImage);
		Point center = new Point(segmentImage.width() / 2, segmentImage.height() / 2);

		Mat rotMat = Imgproc.getRotationMatrix2D(center, degAngle, 1);
		Mat rotated = new Mat();
		Imgproc.warpAffine(segmentImage, rotated, rotMat, segmentImage.size(), Imgproc.CV_WARP_FILL_OUTLIERS,
				Imgproc.BORDER_CONSTANT, avgBgd);

		BufferedImage buffImage = ImageProcessor.mat2Img(rotated);
		TessAPI api = LoadLibs.getTessAPIInstance();
		TessBaseAPI handle = TessAPI1.TessBaseAPICreate();
		String datapath = System.getenv("TESSDATA_PREFIX");
		if (datapath.isEmpty()) {
			datapath = "Tesseract/";
		}
		String language = "eng";

		ByteBuffer buf = ImageIOHelper.convertImageData(buffImage);
		int bpp = buffImage.getColorModel().getPixelSize();
		int bytespp = bpp / 8;
		int bytespl = (int) Math.ceil(buffImage.getWidth() * bpp / 8.0);
		api.TessBaseAPIInit3(handle, datapath, language);
		api.TessBaseAPISetPageSegMode(handle, TessPageSegMode.PSM_AUTO_OSD);

		TessAPI1.TessBaseAPIInit3(handle, datapath, language);
		TessAPI1.TessBaseAPISetPageSegMode(handle, TessAPI1.TessPageSegMode.PSM_AUTO_ONLY);
		TessAPI1.TessBaseAPISetImage(handle, buf, buffImage.getWidth(), buffImage.getHeight(), bytespp, bytespl);
		TessAPI1.TessBaseAPIGetConnectedComponents(handle, null);
		TessAPI1.TessResultIterator ri = TessAPI1.TessBaseAPIGetIterator(handle);
		TessPageIterator pi = TessAPI1.TessResultIteratorGetPageIterator(ri);

		if (pi == null) {
			return null;
		}

		TessAPI1.TessPageIteratorBegin(pi);

		if (rotated.depth() > 1) {
			Imgproc.cvtColor(rotated, rotated, Imgproc.COLOR_BGR2GRAY);
		}

		do {
			IntBuffer leftB = IntBuffer.allocate(1);
			IntBuffer topB = IntBuffer.allocate(1);
			IntBuffer rightB = IntBuffer.allocate(1);
			IntBuffer bottomB = IntBuffer.allocate(1);

			TessAPI1.TessPageIteratorBoundingBox(pi, TessAPI1.TessPageIteratorLevel.RIL_TEXTLINE, leftB, topB, rightB,
					bottomB);

			int left = leftB.get();
			int top = topB.get();
			int right = rightB.get();
			int bottom = bottomB.get();

			Point topLeft = new Point(left, top);
			Point topRight = new Point(right, top);
			Point bottomRight = new Point(right, bottom);
			Point bottomLeft = new Point(left, bottom);

			MatOfPoint line = new MatOfPoint(topLeft, topRight, bottomRight, bottomLeft);
			lines.add(line);

			// Rect rect = Imgproc.boundingRect(line);
			// Mat image = new Mat(rotated, rect);
			// Highgui.imwrite("C:/Users/Reul/Desktop/ZeilenSeg/" + cnt +
			// ".png", image);
			// cnt++;

		} while (TessAPI1.TessPageIteratorNext(pi, TessAPI1.TessPageIteratorLevel.RIL_TEXTLINE) == TessAPI1.TRUE);

		TessAPI1.TessBaseAPIDelete(handle);
		
		ImageProcessor.releaseAll(rotated, rotMat);

		if (postProcess) {
			ArrayList<MatOfPoint> processedlines = LinePostProc.postProcessLines(lines);
			return reverseAngle(processedlines, angle, center);
		}

		return reverseAngle(lines, angle, center);
	}

	public ArrayList<MatOfPoint> reverseAngle(ArrayList<MatOfPoint> lines, double angle, Point center) {
		ArrayList<MatOfPoint> correctedLines = new ArrayList<MatOfPoint>();

		double x_C = center.x;
		double y_C = center.y;

		for (MatOfPoint line : lines) {
			Point[] points = line.toArray();

			for (int i = 0; i < points.length; i++) {
				double x_1 = points[i].x;
				double y_1 = points[i].y;

				double dx = x_1 - x_C;
				double dy = y_1 - y_C;

				double sin = Math.sin(angle);
				double cos = Math.cos(angle);

				double x_2 = (dx * cos) - (dy * sin) + x_C;
				double y_2 = (dx * sin) + (dy * cos) + y_C;

				points[i] = new Point(x_2, y_2);
			}

			MatOfPoint correctedLine = new MatOfPoint(points);
			correctedLines.add(correctedLine);
		}

		return correctedLines;
	}

	public float determineAngle(Mat segment) {
		TessAPI api = LoadLibs.getTessAPIInstance();
		TessBaseAPI handle = TessAPI1.TessBaseAPICreate();

		String datapath = System.getenv("TESSDATA_PREFIX");
		if (datapath.isEmpty()) {
			datapath = "Tesseract/";
		}
		// String datapath = "C:/Users/Reul/workspace/TesseractTest/";
		String language = "eng";

		IntBuffer orientation = IntBuffer.allocate(1);
		IntBuffer direction = IntBuffer.allocate(1);
		IntBuffer order = IntBuffer.allocate(1);
		FloatBuffer deskew_angle = FloatBuffer.allocate(1);

		BufferedImage buffImage = ImageProcessor.mat2Img(segment);

		ByteBuffer buf = ImageIOHelper.convertImageData(buffImage);
		int bpp = buffImage.getColorModel().getPixelSize();
		int bytespp = bpp / 8;
		int bytespl = (int) Math.ceil(buffImage.getWidth() * bpp / 8.0);
		api.TessBaseAPIInit3(handle, datapath, language);
		api.TessBaseAPISetPageSegMode(handle, TessPageSegMode.PSM_AUTO_OSD);

		api.TessBaseAPISetImage(handle, buf, buffImage.getWidth(), buffImage.getHeight(), bytespp, bytespl);
		TessPageIterator pi = api.TessBaseAPIAnalyseLayout(handle);

		if (pi == null) {
			return (float) -360.0;
		}

		try {
			api.TessPageIteratorOrientation(pi, orientation, direction, order, deskew_angle);
		} catch (Exception e) {
			return (float) -360.0;
		}

		float angleRad = deskew_angle.get();
		
		TessAPI1.TessBaseAPIDelete(handle);
		
		return angleRad;
	}

	public Mat extractSegment(MatOfPoint polygon) {
		Mat segmentImage = new Mat(image.size(), image.type(), avgBgd);
		// Mat segmentImage = new Mat(image.size(), image.type(), new
		// Scalar(255, 255, 255));
		Mat mask = new Mat(image.size(), image.type(), new Scalar(0, 0, 0));

		ArrayList<MatOfPoint> points = new ArrayList<MatOfPoint>();
		points.add(polygon);
		Imgproc.drawContours(mask, points, -1, new Scalar(255, 255, 255), -1);
		image.copyTo(segmentImage, mask);

		ImageProcessor.releaseAll(mask);

		return segmentImage;
	}

	public Page getPage() {
		return page;
	}

	public void setPage(Page page) {
		this.page = page;
	}

	public Mat getImage() {
		return image;
	}

	public void setImage(Mat image) {
		this.image = image;
	}

	public Scalar getAvgBgd() {
		return avgBgd;
	}

	public void setAvgBgd(Scalar avgBgd) {
		this.avgBgd = avgBgd;
	}
}