package linePP.drawing;

import java.util.ArrayList;

import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfPoint;
import org.opencv.core.Point;
import org.opencv.core.Rect;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.imgproc.Imgproc;

import imageProcessing.Contour;
import linePP.data.Line;
import linePP.data.LinePage;
import linePP.data.Segment;
import linePP.helper.Bresenham;
import linePP.helper.ImageProcessor;
import regions.type.RegionType;

public class Cutter {

//	private static int cnt = 0;
	
	private static final double EXTEND_LINE_FACTOR = 0.2;
	private static final int NUMBER_OF_CUT_SEGMENTS = 2;

	private static void pruneContours(ArrayList<MatOfPoint> contours, int desiredNumber) {
		while (contours.size() > desiredNumber) {
			double minSize = Double.MAX_VALUE;
			int minIndex = -1;

			for (int i = 0; i < contours.size(); i++) {
				double area = Imgproc.contourArea(contours.get(i));

				if (area < minSize) {
					minSize = area;
					minIndex = i;
				}
			}

			contours.remove(minIndex);
		}
	}

	private static void drawLinePoints(Mat binary, ArrayList<Point> completedLinePoints) {
		double[] black = { 0 };

		for (int i = 0; i < completedLinePoints.size(); i++) {
			int x = (int) completedLinePoints.get(i).x;
			int y = (int) completedLinePoints.get(i).y;

			binary.put(y, x + 1, black);
		}
	}

	private static ArrayList<Point> completeLines(Mat binary, ArrayList<Point> linePoints, boolean useBottomLine) {
		if (useBottomLine) {
			linePoints = Bresenham.calcPointsOnLine(linePoints.get(2), linePoints.get(3));
		} else {
			linePoints = Bresenham.calcPointsOnLine(linePoints.get(0), linePoints.get(1));
		}

		ArrayList<Point> completedLinePoints = new ArrayList<Point>();
		completedLinePoints.addAll(linePoints);

		for (int i = 0; i < linePoints.size(); i++) {
			int x = (int) linePoints.get(i).x;
			int y = (int) linePoints.get(i).y;

			if (x + 1 < binary.width() && binary.get(y, x + 1)[0] > 0) {
				Point newPoint = new Point(x + 1, y);
				completedLinePoints.add(newPoint);
			}
		}

		return completedLinePoints;
	}

	private static Point calcExtendedLinePoint(Point start, Point end) {
		double lenSE = Math.sqrt(Math.pow((start.x - end.x), 2) + Math.pow((start.y - end.y), 2));

		double x = end.x + (end.x - start.x) / lenSE * (EXTEND_LINE_FACTOR * lenSE);
		double y = end.y + (end.y - start.y) / lenSE * (EXTEND_LINE_FACTOR * lenSE);

		return new Point(x, y);
	}

	private static ArrayList<Point> calcPointsToCheck(Mat image, Line line) {
		Point[] linePoints = ImageProcessor.findLinePointOrder(line.getLinePoints());
		Point topLeft = linePoints[0];
		Point topRight = linePoints[1];
		Point bottomRight = linePoints[2];
		Point bottomLeft = linePoints[3];

		Point extendedPoint0 = calcExtendedLinePoint(topLeft, topRight);
		Point extendedPoint1 = calcExtendedLinePoint(topRight, topLeft);
		Point extendedPoint2 = calcExtendedLinePoint(bottomRight, bottomLeft);
		Point extendedPoint3 = calcExtendedLinePoint(bottomLeft, bottomRight);

		ArrayList<Point> points = new ArrayList<Point>();
		points.add(extendedPoint0);
		points.add(extendedPoint1);
		points.add(extendedPoint2);
		points.add(extendedPoint3);

		return points;
	}

	private static boolean isTopLine(Line line, Segment segment) {
		if (segment.getLines().get(0).equals(line)) {
			return true;
		}

		return false;
	}

	private static int findNextLineIndex(Segment segment, Line line) {
		ArrayList<Line> lines = segment.getLines();

		for (int i = 0; i < lines.size(); i++) {
			if (lines.get(i).equals(line)) {
				if (i + 1 < lines.size()) {
					return i + 1;
				} else {
					return -1;
				}
			}
		}

		return -1;
	}

	private static MatOfPoint chooseContour(ArrayList<MatOfPoint> contours, boolean chooseBottom) {
		if(contours.size() == 1) {
			return contours.get(0);
		}

		Rect rect0 = Imgproc.boundingRect(contours.get(0));
		Rect rect1 = Imgproc.boundingRect(contours.get(1));

		if (chooseBottom) {
			if (rect0.tl().y > rect1.tl().y) {
				return contours.get(1);
			} else {
				return contours.get(0);
			}
		} else {
			if (rect0.tl().y > rect1.tl().y) {
				return contours.get(0);
			} else {
				return contours.get(1);
			}
		}
	}

	private static ArrayList<Line> addLines(Segment oldSegment, Segment newSegment, int lineIndex, boolean addAbove) {
		ArrayList<Line> lines = new ArrayList<Line>();

		if (addAbove) {
			for (int i = 0; i < oldSegment.getLines().size(); i++) {
				if (i < lineIndex) {
					lines.add(oldSegment.getLines().get(i));
				}
			}
		} else {
			for (int i = 0; i < oldSegment.getLines().size(); i++) {
				if (i >= lineIndex) {
					lines.add(oldSegment.getLines().get(i));
				}
			}
		}

		return lines;
	}

	private static MatOfPoint cutSegment(Mat binary, Line line, boolean cutBottom, int expectedContours) {
		ArrayList<Point> linePoints = calcPointsToCheck(binary, line);
		ArrayList<Point> completedLinePoints = completeLines(binary, linePoints, cutBottom);
		
		drawLinePoints(binary, completedLinePoints);

		Mat binaryTemp = binary.clone();
		ArrayList<MatOfPoint> contours = new ArrayList<MatOfPoint>();
		Imgproc.findContours(binaryTemp, contours, new Mat(), Imgproc.RETR_EXTERNAL, Imgproc.CHAIN_APPROX_SIMPLE);
		pruneContours(contours, expectedContours);

		MatOfPoint contour = chooseContour(contours, cutBottom);

		return contour;
	}

	private static ArrayList<Segment> cutLine(LinePage page, Segment segment, Line line, boolean separateLine) {
		int nextLineIndex = findNextLineIndex(segment, line);

		if (nextLineIndex == -1) {
			return null;
		}

		Mat image = page.getOriginal();
		MatOfPoint region = segment.getRegion();
		Mat binary = new Mat(image.size(), CvType.CV_8UC1, new Scalar(0));
		ArrayList<MatOfPoint> contours = new ArrayList<MatOfPoint>();
		contours.add(region);
		Imgproc.drawContours(binary, contours, -1, new Scalar(255), -1);
		
		Line nextLine = segment.getLines().get(nextLineIndex);
		ArrayList<Segment> segments = new ArrayList<Segment>();

		Segment topSegment = null;

		if (separateLine) {
			topSegment = new Segment(cutSegment(binary, segment.getLines().get(nextLineIndex - 1), true, NUMBER_OF_CUT_SEGMENTS),
					segment.getType());
			ArrayList<Line> topLines = addLines(segment, topSegment, nextLineIndex - 1, true);
			topSegment.setLines(topLines);

			Segment lineSegment = new Segment(line.getLinePoints(), RegionType.heading);
			lineSegment.addLine(line);
			segments.add(lineSegment);

			if(!isTopLine(line, segment)) {
				segments.add(topSegment);
			}
		} else {
			topSegment = new Segment(cutSegment(binary, nextLine, true, NUMBER_OF_CUT_SEGMENTS), segment.getType());
			ArrayList<Line> topLines = addLines(segment, topSegment, nextLineIndex, true);
			topSegment.setLines(topLines);
			segments.add(topSegment);
		}
		
		Segment bottomSegment = new Segment(cutSegment(binary, line, false, NUMBER_OF_CUT_SEGMENTS), segment.getType());
		ArrayList<Line> bottomLines = addLines(segment, bottomSegment, nextLineIndex, false);
		bottomSegment.setLines(bottomLines);
		segments.add(bottomSegment);

		return segments;
	}

	private static void pruneSegment(Segment segment, Mat image) {
		Mat binary = new Mat(image.size(), CvType.CV_8UC1, new Scalar(0));
		ArrayList<MatOfPoint> contours = new ArrayList<MatOfPoint>();
		contours.add(segment.getRegion());
		Imgproc.drawContours(binary, contours, -1, new Scalar(255), -1);
		
		Imgproc.morphologyEx(binary, binary, Imgproc.MORPH_CLOSE, Mat.ones(new Size(3, 3), CvType.CV_8U));
		contours = Contour.findContours(binary.clone());
		
		if(contours.size() > 0) {
			segment.setRegion(contours.get(0));
		}
	}

	public static ArrayList<Segment> performCut(LinePage page, Segment segment, Line line, boolean separateLine) {
		ArrayList<Segment> newSegments = cutLine(page, segment, line, separateLine);

		if (newSegments == null) {
			return null;
		}

		for(Segment tempSegment : newSegments) {
			if(!tempSegment.isPruned()) {
				pruneSegment(tempSegment, page.getOriginal());
				tempSegment.setPruned(true);
				tempSegment.resizeAll(page.getScaleFactor());
			}
			
			tempSegment.setLeft(segment.isLeft());
		}

		page.getSegments().remove(segment);
		page.getSegments().addAll(newSegments);
		page.update();
		
		return newSegments;
	}
}