package linePP.drawing;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.MouseInfo;
import java.awt.RenderingHints;
import java.awt.event.MouseEvent;

import javax.swing.SwingUtilities;

import org.opencv.core.Point;

import linePP.auto.LineType;
import linePP.data.Line;
import linePP.data.Polygon;
import linePP.data.Segment;
import linePP.helper.ImageProcessor;
import regions.RegionManager;
import regions.type.RegionType;

public class Painter {
	private static final int THICKNESS = 2;

	private BackgroundPanel panel;
	private LineRegionManager regionManager;

	private Graphics g;
	private Graphics2D g2D;

	private Segment activeSegment;
	private Line activeLine;

	private boolean assigningMarginalia;
	private boolean showMarginalia;

	private java.awt.Point start;
	private java.awt.Point end;

	public Painter(BackgroundPanel panel) {
		setPanel(panel);
	}

	public void reset() {
		setActiveSegment(null);
		setActiveLine(null);
	}

	public void fullReset() {
		setActiveSegment(null);
		setActiveLine(null);
		setStart(null);
		setEnd(null);
		setAssigningMarginalia(false);
	}

	public void handleMouseMoved(MouseEvent e) {
		Point clicked = new Point(e.getX(), e.getY());
		Segment clickedSegment = regionManager.identifyClickedSegment(clicked);

		if (clickedSegment != null) {
			if (clickedSegment.equals(activeSegment)) {
				Line clickedLine = clickedSegment.identifyClickedLine(clicked);

				if (clickedLine != null) {
					setActiveLine(clickedLine);
				}
			} else {
				setActiveSegment(clickedSegment);
			}
		} else {
			reset();
		}

		if (assigningMarginalia) {
			Point point = new Point(e.getX(), e.getY());
			Segment segment = regionManager.identifyClickedSegment(point);

			if (segment != null) {
				if (segment.getType().equals(RegionType.marginalia)) {
					Point cog = ImageProcessor.calcCenterOfGravity(segment.getResizedRegion());
					setEnd(new java.awt.Point((int) cog.x, (int) cog.y));
				}
			} else {
				setEnd(null);
			}
		}
	}

	public void handleMousePressed(MouseEvent e) {
		if (SwingUtilities.isLeftMouseButton(e)) {
			Point clicked = new Point(e.getX(), e.getY());
			Segment clickedSegment = regionManager.identifyClickedSegment(clicked);

			if (clickedSegment == null) {
				fullReset();
			}
		} else { // right click
			if (activeLine != null) {
				new PopUp(MouseInfo.getPointerInfo().getLocation(), this);
			}
		}
	}

	private void drawPolygon(Polygon polygon, RegionType type, boolean fill) {
		if (type == null)
			System.out.println("type null");

		Color color = RegionManager.getColorByRegionType(type);

		if (color != null) {
			g.setColor(color);
			g.drawPolygon(polygon.getX_coords(), polygon.getY_coords(), polygon.getN());

			if (fill) {
				g.setColor(new Color(color.getRed(), color.getGreen(), color.getBlue(), 120));
				g.fillPolygon(polygon.getX_coords(), polygon.getY_coords(), polygon.getN());
			}
		} else {
			System.out.println("Color = null!");
		}
	}

	public void paint(Graphics g) {
		setG(g);
		setG2D((Graphics2D) g);

		// paints the background image
		g.drawImage(panel.getBuffImage(), 0, 0, null);
		g2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);

		// paint segments and lines
		g2D.setStroke(new BasicStroke(THICKNESS));

		for (Segment segment : regionManager.getPage().getSegments()) {
			if (!segment.isActive()) {
				drawPolygon(segment.getPolyRegion(), segment.getType(), false);
			} else {
				for (Line line : segment.getLines()) {
					if (line.getLineType().equals(LineType.HIGH)) {

					} else {

						if (line.isActive()) {
							drawPolygon(line.getResizedLine().getPolygon(), segment.getType(), true);
						} else {
							drawPolygon(line.getResizedLine().getPolygon(), segment.getType(), false);
						}
					}
				}
			}

			for (Line line : segment.getLines()) {
				if (line.getLineType().equals(LineType.HIGH)) {
					drawPolygon(line.getResizedLine().getPolygon(), RegionType.ignore, true);
				} else if (line.getLineType().equals(LineType.BELOW_HIGH)) {
					drawPolygon(line.getResizedLine().getPolygon(), RegionType.heading, true);
				} else if (line.getLineType().equals(LineType.HIGH_CANDIDATE)) {
					drawPolygon(line.getResizedLine().getPolygon(), RegionType.paragraph, true);
				} else if (line.getLineType().equals(LineType.HEADING)) {
					drawPolygon(line.getResizedLine().getPolygon(), RegionType.signature_mark, true);
				}
			}

			if (assigningMarginalia && segment.getType().equals(RegionType.marginalia)) {
				drawPolygon(segment.getPolyRegion(), segment.getType(), true);
			}

			if (showMarginalia && segment.getType().equals(RegionType.marginalia) && segment.getMarginalia() != null) {
				drawMarginaliaConnector(segment.getMarginalia().getStart(), segment.getMarginalia().getEnd());
			}
		}

		if (assigningMarginalia) {
			if (start != null && end != null) {
				drawMarginaliaConnector(start, end);
			}
		}
	}

	private void drawMarginaliaConnector(java.awt.Point start, java.awt.Point end) {
		g2D.setColor(Color.ORANGE);
		g2D.setStroke(new BasicStroke(THICKNESS * 3));
		g2D.drawLine(start.x, start.y, end.x, end.y);
		g2D.setStroke(new BasicStroke(THICKNESS));
	}

	public BackgroundPanel getPanel() {
		return panel;
	}

	public void setPanel(BackgroundPanel panel) {
		this.panel = panel;
	}

	public LineRegionManager getRegionManager() {
		return regionManager;
	}

	public void setRegionManager(LineRegionManager regionManager) {
		this.regionManager = regionManager;
	}

	public Graphics getG() {
		return g;
	}

	public void setG(Graphics g) {
		this.g = g;
	}

	public Graphics2D getG2D() {
		return g2D;
	}

	public void setG2D(Graphics2D g2d) {
		g2D = g2d;
	}

	public Segment getActiveSegment() {
		return activeSegment;
	}

	public void setActiveSegment(Segment activeSegment) {
		if (this.activeSegment != null) {
			this.activeSegment.setActive(false);
		}
		if (activeSegment != null) {
			activeSegment.setActive(true);
		}
		this.activeSegment = activeSegment;
	}

	public Line getActiveLine() {
		return activeLine;
	}

	public void setActiveLine(Line activeLine) {
		if (this.activeLine != null) {
			this.activeLine.setActive(false);
		}
		if (activeLine != null) {
			activeLine.setActive(true);
		}
		this.activeLine = activeLine;
	}

	public boolean isAssigningMarginalia() {
		return assigningMarginalia;
	}

	public void setAssigningMarginalia(boolean assigningMarginalia) {
		this.assigningMarginalia = assigningMarginalia;
	}

	public boolean isShowMarginalia() {
		return showMarginalia;
	}

	public void setShowMarginalia(boolean showMarginalia) {
		this.showMarginalia = showMarginalia;
	}

	public java.awt.Point getStart() {
		return start;
	}

	public void setStart(java.awt.Point start) {
		this.start = start;
	}

	public java.awt.Point getEnd() {
		return end;
	}

	public void setEnd(java.awt.Point end) {
		this.end = end;
	}
}