package segmentation.result;

import java.awt.Color;
import java.util.ArrayList;

import org.opencv.core.Core;
import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfPoint;
import org.opencv.core.Point;
import org.opencv.core.Rect;
import org.opencv.core.Scalar;
import org.opencv.imgproc.Imgproc;

import geometry.Polygon;
import regions.RegionManager;
import regions.type.RegionType;
import segmentation.ImageProcessor;

@SuppressWarnings("rawtypes")
public class ResultRegion implements Comparable {

	private RegionType type;
	private Scalar color;
	private int imageHeight;
	private MatOfPoint points;

	private Rect rect;
	private MatOfPoint scaledPoints;
	private int readingOrderIndex;
	private boolean isAssigned;
	private boolean isActive;
	private Polygon scaledPolygon;
	private java.awt.Point scaledCenterOfGravity;

	private Mat roBinary;
	private ArrayList<Point> containedPoints;

	public ResultRegion(RegionType type, int imageHeight, MatOfPoint points) {
		setType(type);
		Scalar color = RegionManager.getScalarByRegionType(type);
		setColor(color);
		setImageHeight(imageHeight);
		setPoints(points);
		setReadingOrderIndex(-1);
	}

	public void calcROBinary(Mat image) {
		Mat binary = new Mat(image.size(), CvType.CV_8UC1, new Scalar(0));
		Core.fillConvexPoly(binary, scaledPoints, new Scalar(255));
		setRoBinary(binary);

		Rect rect = Imgproc.boundingRect(scaledPoints);
		ArrayList<Point> containedPoints = new ArrayList<Point>();

		for (int y = rect.y; y < rect.y + rect.height; y++) {
			for (int x = rect.x; x < rect.x + rect.width; x++) {
				if (binary.get(y, x)[0] > 0) {
					containedPoints.add(new Point(x, y));
				}
			}
		}

		setContainedPoints(containedPoints);
	}

	public void applyScaleCorrection(Mat original, Mat resized) {
		double scaleFactor = (double) resized.height() / imageHeight;
		Point[] originalPoints = points.toArray();
		Point[] scaledPointsTemp = new Point[originalPoints.length];

		for (int i = 0; i < originalPoints.length; i++) {
			Point point = new Point(originalPoints[i].x * scaleFactor, originalPoints[i].y * scaleFactor);
			scaledPointsTemp[i] = point;
		}

		MatOfPoint scaledPoints = new MatOfPoint(scaledPointsTemp);
		ArrayList<MatOfPoint> contours = new ArrayList<MatOfPoint>();
		contours.add(scaledPoints);

		Mat binaryMat = new Mat(resized.size(), CvType.CV_8U, new Scalar(0));
		Imgproc.drawContours(binaryMat, contours, -1, new Scalar(255), -1);
		Imgproc.findContours(binaryMat, contours, new Mat(), Imgproc.RETR_EXTERNAL, Imgproc.CHAIN_APPROX_SIMPLE);

		setScaledPoints(contours.get(0));
		setScaledPolygon(new Polygon(contours.get(0)));
		setScaledCenterOfGravity(ImageProcessor.calcCenterOfGravity(scaledPoints));
	}

	public MatOfPoint scalePointsSimple(double scaleFactor) {
		Point[] originalPoints = points.toArray();
		Point[] scaledPointsTemp = new Point[originalPoints.length];

		for (int i = 0; i < originalPoints.length; i++) {
			Point point = new Point(originalPoints[i].x * scaleFactor, originalPoints[i].y * scaleFactor);
			scaledPointsTemp[i] = point;
		}

		MatOfPoint scaledPoints = new MatOfPoint(scaledPointsTemp);

		return scaledPoints;
	}

	public RegionType getType() {
		return type;
	}

	public void setType(RegionType type) {
		this.type = type;

		Color color = RegionManager.getColorByRegionType(type);
		if(color != null) {
			setColor(new Scalar(color.getBlue(), color.getGreen(), color.getRed()));
		} else {
			System.out.println("Color for type " + type.toString() + " undefined. Please check Settings.");
			setColor(new Scalar(0, 0, 0));
		}
	}

	public Scalar getColor() {
		return color;
	}

	public void setColor(Scalar color) {
		this.color = color;
	}

	public int getImageHeight() {
		return imageHeight;
	}

	public void setImageHeight(int imageHeight) {
		this.imageHeight = imageHeight;
	}

	public MatOfPoint getPoints() {
		return points;
	}

	public void setPoints(MatOfPoint points) {
		setRect(Imgproc.boundingRect(points));
		this.points = points;
	}

	public MatOfPoint getScaledPoints() {
		return scaledPoints;
	}

	public void setScaledPoints(MatOfPoint scaledPoints) {
		this.scaledPoints = scaledPoints;
	}

	public Polygon getScaledPolygon() {
		return scaledPolygon;
	}

	public void setScaledPolygon(Polygon scaledPolygon) {
		this.scaledPolygon = scaledPolygon;
	}

	public int getReadingOrderIndex() {
		return readingOrderIndex;
	}

	public void setReadingOrderIndex(int readingOrderIndex) {
		this.readingOrderIndex = readingOrderIndex;
	}

	public boolean isActive() {
		return isActive;
	}

	public void setActive(boolean isActive) {
		this.isActive = isActive;
	}

	public boolean isAssigned() {
		return isAssigned;
	}

	public void setAssigned(boolean isAssigned) {
		this.isAssigned = isAssigned;
	}

	public java.awt.Point getScaledCenterOfGravity() {
		return scaledCenterOfGravity;
	}

	public void setScaledCenterOfGravity(java.awt.Point scaledCenterOfGravity) {
		this.scaledCenterOfGravity = scaledCenterOfGravity;
	}

	public Rect getRect() {
		return rect;
	}

	public void setRect(Rect rect) {
		this.rect = rect;
	}

	public int compareTo(Object o) {
		ResultRegion toCompare = (ResultRegion) o;

		if (toCompare.getRect().y < this.getRect().y) {
			return 1;
		} else {
			return -1;
		}
	}

	public Mat getRoBinary() {
		return roBinary;
	}

	public void setRoBinary(Mat roBinary) {
		this.roBinary = roBinary;
	}

	public ArrayList<Point> getContainedPoints() {
		return containedPoints;
	}

	public void setContainedPoints(ArrayList<Point> containedPoints) {
		this.containedPoints = containedPoints;
	}
}